from typing import Union

from udata_hydra import context
from udata_hydra.db import (
    compute_insert_query,
    convert_dict_values_to_json,
    update_table_record,
)


class Check:
    """Represents a check in the "checks" DB table"""

    @classmethod
    async def get(cls, check_id: int) -> dict:
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = """
                SELECT * FROM catalog JOIN checks
                ON catalog.last_check = checks.id
                WHERE checks.id = $1;
            """
            check = await connection.fetchrow(q, check_id)
        return check

    @classmethod
    async def get_latest(
        cls, url: Union[str, None], resource_id: Union[str, None]
    ) -> Union[dict, None]:
        column: str = "url" if url else "resource_id"
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = f"""
            SELECT catalog.id as catalog_id, checks.id as check_id,
                catalog.status as catalog_status, checks.status as check_status, catalog.deleted as deleted, *
            FROM checks, catalog
            WHERE checks.id = catalog.last_check
            AND catalog.{column} = $1
            """
            return await connection.fetchrow(q, url or resource_id)

    @classmethod
    async def get_all(
        cls, url: Union[str, None], resource_id: Union[str, None]
    ) -> Union[dict, None]:
        column: str = "url" if url else "resource_id"
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = f"""
            SELECT catalog.id as catalog_id, checks.id as check_id,
                catalog.status as catalog_status, checks.status as check_status, catalog.deleted as deleted, *
            FROM checks, catalog
            WHERE catalog.{column} = $1
            AND catalog.url = checks.url
            ORDER BY created_at DESC
            """
            return await connection.fetch(q, url or resource_id)

    @classmethod
    async def insert(cls, data: dict) -> int:
        """
        Insert a new check in DB and return the check id in DB
        This use the info from the last check of the same resource
        """
        data = convert_dict_values_to_json(data)
        q = compute_insert_query(table_name="checks", data=data)
        pool = await context.pool()
        async with pool.acquire() as connection:
            last_check = await connection.fetchrow(q, *data.values())
            q = """UPDATE catalog SET last_check = $1 WHERE resource_id = $2"""
            await connection.execute(q, last_check["id"], data["resource_id"])
        return last_check["id"]

    @classmethod
    async def update(cls, check_id: int, data: dict) -> int:
        """Update a check in DB with new data and return the check id in DB"""
        return await update_table_record(table_name="checks", record_id=check_id, data=data)

    @classmethod
    async def delete(cls, check_id: int) -> int:
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = """DELETE FROM checks WHERE id = $1"""
            return await connection.fetch(q, check_id)



from udata_hydra import context


class Resource:
    """Represents a resource in the "catalog" DB table"""

    @classmethod
    async def get(cls, resource_id: str, column_name: str = "*") -> dict:
        pool = await context.pool()
        async with pool.acquire() as connection:
            q = f"""SELECT {column_name} FROM catalog WHERE resource_id = '{resource_id}';"""
            resource = await connection.fetchrow(q)
        return resource

    @classmethod
    async def insert(
        cls,
        dataset_id: str,
        resource_id: str,
        url: str,
        priority: bool = True,
    ) -> None:
        pool = await context.pool()
        async with pool.acquire() as connection:
            # Insert new resource in catalog table and mark as high priority for crawling
            q = f"""
                    INSERT INTO catalog (dataset_id, resource_id, url, deleted, status, priority)
                    VALUES ('{dataset_id}', '{resource_id}', '{url}', FALSE, 'TO_CHECK', '{priority}')
                    ON CONFLICT (resource_id) DO UPDATE SET
                        dataset_id = '{dataset_id}',
                        url = '{url}',
                        priority = '{priority}';"""
            await connection.execute(q)

    @classmethod
    async def update(cls, resource_id: str, data: dict) -> str:
        """Update a resource in DB with new data and return the updated resource id in DB"""
        columns = data.keys()
        # $1, $2...
        placeholders = [f"${x + 1}" for x in range(len(data.values()))]
        set_clause = ",".join([f"{c} = {v}" for c, v in zip(columns, placeholders)])
        q = f"""
                UPDATE catalog
                SET {set_clause}
                WHERE resource_id = ${len(placeholders) + 1};"""
        pool = await context.pool()
        await pool.execute(q, *data.values(), resource_id)
        return resource_id

    @classmethod
    async def update_or_insert(
        cls,
        dataset_id: str,
        resource_id: str,
        url: str,
        priority: bool = True,  # Make resource high priority by default for crawling
    ) -> None:
        pool = await context.pool()
        async with pool.acquire() as connection:
            # Check if resource is in catalog then insert or update into table
            if await Resource.get(resource_id):
                q = f"""
                        UPDATE catalog
                        SET url = '{url}', priority = '{priority}'
                        WHERE resource_id = '{resource_id}';"""
            else:
                q = f"""
                        INSERT INTO catalog (dataset_id, resource_id, url, deleted, priority)
                        VALUES ('{dataset_id}', '{resource_id}', '{url}', FALSE, '{priority}')
                        ON CONFLICT (resource_id) DO UPDATE SET
                            dataset_id = '{dataset_id}',
                            url = '{url}',
                            priority = '{priority}';"""
            await connection.execute(q)



from typing import Generator

import pyarrow as pa
import pyarrow.parquet as pq

PYTHON_TYPE_TO_PA = {
    "string": pa.string(),
    "float": pa.float64(),
    "int": pa.int64(),
    "bool": pa.bool_(),
    "json": pa.string(),
    "date": pa.date32(),
    "datetime": pa.date64(),
}


def save_as_parquet(
    records: Generator,
    columns: dict,
    output_name: str,
    save_output: bool = True,
) -> tuple[str, pa.Table]:
    # the "save_output" argument is only used in tests
    table = pa.Table.from_pylist(
        [{c: v for c, v in zip(columns, values)} for values in records],
        schema=pa.schema([pa.field(c, PYTHON_TYPE_TO_PA[columns[c]]) for c in columns]),
    )
    if save_output:
        pq.write_table(table, f"{output_name}.parquet")
    return f"{output_name}.parquet", table



from aiohttp import web

from udata_hydra.db.check import Check
from udata_hydra.schemas import CheckSchema
from udata_hydra.utils import get_request_params


async def get_latest_check(request: web.Request) -> web.Response:
    """Get the latest check for a given URL or resource_id"""
    url, resource_id = get_request_params(request, params_names=["url", "resource_id"])
    data = await Check.get_latest(url, resource_id)
    if not data:
        raise web.HTTPNotFound()
    if data["deleted"]:
        raise web.HTTPGone()
    return web.json_response(CheckSchema().dump(dict(data)))


async def get_all_checks(request: web.Request) -> web.Response:
    url, resource_id = get_request_params(request, params_names=["url", "resource_id"])
    data = await Check.get_all(url, resource_id)
    if not data:
        raise web.HTTPNotFound()
    return web.json_response([CheckSchema().dump(dict(r)) for r in data])
